#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple, Callable
from logging import Logger
import pickle
import pandas as pd
from sklearn.gaussian_process.kernels import RBF, DotProduct
from mgktools.kernels.utils import get_kernel_config
from mgktools.data.data import Dataset


class EmptyLogger:
    def debug(self, info):
        return

    def info(self, info):
        return


def get_data(data_format: Literal['mgktools', 'chemprop', 'fingerprints'],
             path: str,
             pure_columns: List[str] = None,
             mixture_columns: List[str] = None,
             target_columns: List[str] = None,
             feature_columns: List[str] = None,
             features_generator: List[Union[str, Callable]] = None,
             features_combination: Literal['concat', 'mean'] = None,
             graph_kernel_type: Literal['graph', 'pre-computed'] = None,
             n_jobs: int = 8):
    if data_format == 'fingerprints':
        from molalkit.data.utils import get_data
        dataset = get_data(path=path,
                           pure_columns=pure_columns,
                           mixture_columns=mixture_columns,
                           target_columns=target_columns,
                           feature_columns=feature_columns,
                           features_generator=features_generator,
                           features_combination=features_combination,
                           n_jobs=n_jobs)
    elif data_format == 'chemprop':
        from chemprop.data.utils import get_data
        assert mixture_columns is None
        assert feature_columns is None
        dataset = get_data(path=path,
                           smiles_columns=pure_columns,
                           target_columns=target_columns,
                           features_generator=features_generator)
    elif data_format == 'mgktools':
        assert graph_kernel_type is not None
        from mgktools.data.data import get_data
        dataset = get_data(path=path,
                           pure_columns=pure_columns,
                           mixture_columns=mixture_columns,
                           target_columns=target_columns,
                           feature_columns=feature_columns,
                           features_generator=features_generator,
                           features_combination=features_combination,
                           mixture_type='single_graph',
                           graph_kernel_type=graph_kernel_type,
                           n_jobs=n_jobs)
    else:
        raise ValueError('input error')
    df = pd.read_csv(path)
    if 'id' not in df:
        df['id'] = range(len(df))
    for i, data in enumerate(dataset):
        data.id = df.iloc[i]['id']
    return dataset


def get_model(data_format: Literal['mgktools', 'chemprop', 'fingerprints'],
              dataset_type: Literal['regression', 'classification', 'multiclass'],
              model: Literal['random_forest', 'naive_bayes', 'gaussian_process', 'support_vector_machine'],
              save_dir: str = None,
              loss_function: Literal['mse', 'bounded_mse', 'binary_cross_entropy', 'cross_entropy', 'mcc', 'sid',
              'wasserstein', 'mve', 'evidential', 'dirichlet'] = None,
              num_tasks: int = 1,
              multiclass_num_classes: int = 3,
              features_generator=None,
              no_features_scaling: bool = False,
              features_only: bool = False,
              features_size: int = 0,
              epochs: int = 30,
              depth: int = 3,
              hidden_size: int = 300,
              ffn_num_layers: int = 2,
              ffn_hidden_size: int = None,
              dropout: float = 0.0,
              batch_size: int = 50,
              ensemble_size: int = 1,
              number_of_molecules: int = 1,
              mpn_shared: bool = False,
              atom_messages: bool = False,
              undirected: bool = False,
              class_balance: bool = False,
              checkpoint_dir: str = None,
              checkpoint_frzn: str = None,
              frzn_ffn_layers: int = 0,
              freeze_first_only: bool = False,
              kernel=None,
              uncertainty_type: Literal['value', 'uncertainty'] = None,
              alpha: Union[float, str] = 1e-8,
              n_jobs: int = 8,
              seed: int = 0,
              logger: Logger = None,
              ):
    if alpha.__class__ == str:
        alpha = float(open(alpha).read())

    if data_format == 'fingerprints':
        if model == 'random_forest':
            if dataset_type == 'regression':
                from molalkit.models.random_forest.RandomForestRegressor import RFRegressor
                return RFRegressor(random_state=seed)
            else:
                from molalkit.models.random_forest.RandomForestClassifier import RFClassifier
                return RFClassifier(random_state=seed, oob_score=True)
        elif model == 'naive_bayes':
            assert dataset_type == 'classification'
            from molalkit.models.naive_bayes.NaiveBayesClassifier import NBClassifier
            return NBClassifier()
        elif model == 'logistic_regression':
            assert dataset_type == 'classification'
            from molalkit.models.logistic_regression.LogisticRegression import LogisticRegressor
            return LogisticRegressor(random_state=seed)
        elif model == 'gaussian_process_regression':
            assert dataset_type in ['regression', 'classification']
            assert uncertainty_type is not None
            from molalkit.models.gaussian_process.GaussianProcessRegressor import GPRegressor
            return GPRegressor(kernel=kernel, alpha=alpha, optimizer=None, uncertainty_type=uncertainty_type)
        elif model == 'gaussian_process_classification':
            assert dataset_type == 'classification'
            from molalkit.models.gaussian_process.GaussianProcessClassifier import GPClassifier
            return GPClassifier(kernel=kernel, optimizer=None)
        elif model == 'support_vector_machine':
            assert dataset_type == 'classification'
            from molalkit.models.support_vector.SupportVectorClassifier import SVClassifier
            return SVClassifier(kernel=kernel, probability=True)
        else:
            raise ValueError(f'unknown model: {model}')
    elif data_format == 'chemprop':
        from molalkit.models.mpnn.mpnn import MPNN
        return MPNN(save_dir=save_dir,
                    dataset_type=dataset_type,
                    loss_function=loss_function,
                    num_tasks=num_tasks,
                    multiclass_num_classes=multiclass_num_classes,
                    features_generator=features_generator,
                    no_features_scaling=no_features_scaling,
                    features_only=features_only,
                    features_size=features_size,
                    epochs=epochs,
                    depth=depth,
                    hidden_size=hidden_size,
                    ffn_num_layers=ffn_num_layers,
                    ffn_hidden_size=ffn_hidden_size,
                    dropout=dropout,
                    batch_size=batch_size,
                    ensemble_size=ensemble_size,
                    number_of_molecules=number_of_molecules,
                    mpn_shared=mpn_shared,
                    atom_messages=atom_messages,
                    undirected=undirected,
                    class_balance=class_balance,
                    checkpoint_dir=checkpoint_dir,
                    checkpoint_frzn=checkpoint_frzn,
                    frzn_ffn_layers=frzn_ffn_layers,
                    freeze_first_only=freeze_first_only,
                    n_jobs=n_jobs,
                    seed=seed,
                    logger=EmptyLogger())
    elif data_format == 'mgktools':
        if model == 'gaussian_process_regression':
            assert dataset_type in ['regression', 'classification']
            assert uncertainty_type is not None
            from molalkit.models.gaussian_process.GaussianProcessRegressor import GPRegressor
            return GPRegressor(kernel=kernel, alpha=alpha, optimizer=None, uncertainty_type=uncertainty_type)
        elif model == 'gaussian_process_classification':
            assert dataset_type == 'classification'
            from molalkit.models.gaussian_process.GaussianProcessClassifier import GPClassifier
            return GPClassifier(kernel=kernel, optimizer=None)
        elif model == 'support_vector_machine':
            assert dataset_type == 'classification'
            from molalkit.models.support_vector.SupportVectorClassifier import SVClassifier
            return SVClassifier(kernel=kernel, probability=True)
        else:
            raise ValueError(f'unknown model: {model}')
    else:
        raise ValueError(f'unknown data_format {data_format}')


def get_kernel(graph_kernel_type: Literal['graph', 'pre-computed'] = None,
               mgk_files: List[str] = None,
               features_kernel_type: Literal['linear', 'dot_product', 'rbf'] = None,
               features_hyperparameters: Union[float, List[float]] = None,
               features_hyperparameters_file: str = None,
               dataset: Dataset = None,
               kernel_pkl_path: str = None,
               ):
    if mgk_files is None:
        assert graph_kernel_type is None
        # no graph kernel involved.
        if features_kernel_type is None:
            return None
        elif features_kernel_type == 'linear':
            return 'linear'
        elif features_kernel_type == 'dot_product':
            if features_hyperparameters.__class__ == list:
                assert len(features_hyperparameters) == 1
                sigma_0 = features_hyperparameters[0]
            else:
                sigma_0 = features_hyperparameters
            return DotProduct(sigma_0=sigma_0)
        elif features_kernel_type == 'rbf':
            return RBF(length_scale=features_hyperparameters)
        else:
            raise ValueError
    else:
        if graph_kernel_type == 'graph':
            return get_kernel_config(
                dataset=dataset,
                graph_kernel_type='graph',
                mgk_hyperparameters_files=mgk_files,
                features_kernel_type=features_kernel_type,
                features_hyperparameters=features_hyperparameters,
                features_hyperparameters_bounds="fixed",
                features_hyperparameters_file=features_hyperparameters_file
            ).kernel
        elif graph_kernel_type == 'pre-computed':
            assert kernel_pkl_path is not None
            if os.path.exists(kernel_pkl_path):
                return get_kernel_config(
                    dataset=dataset,
                    graph_kernel_type='pre-computed',
                    features_kernel_type=features_kernel_type,
                    features_hyperparameters=features_hyperparameters,
                    features_hyperparameters_bounds="fixed",
                    features_hyperparameters_file=features_hyperparameters_file,
                    kernel_pkl=kernel_pkl_path
                ).kernel
            else:
                dataset.graph_kernel_type = 'graph'
                kernel_config = get_kernel_config(
                    dataset=dataset,
                    graph_kernel_type='graph',
                    mgk_hyperparameters_files=mgk_files,
                    features_kernel_type=features_kernel_type,
                    features_hyperparameters=features_hyperparameters,
                    features_hyperparameters_bounds="fixed",
                    features_hyperparameters_file=features_hyperparameters_file
                )
                kernel_dict = kernel_config.get_kernel_dict(dataset.X, dataset.X_repr.ravel())
                dataset.graph_kernel_type = 'pre-computed'
                pickle.dump(kernel_dict, open(kernel_pkl_path, 'wb'), protocol=4)
                return get_kernel_config(
                    dataset=dataset,
                    graph_kernel_type='pre-computed',
                    features_kernel_type=features_kernel_type,
                    features_hyperparameters=features_hyperparameters,
                    features_hyperparameters_bounds="fixed",
                    features_hyperparameters_file=features_hyperparameters_file,
                    kernel_dict=kernel_dict
                ).kernel
        else:
            raise ValueError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble._forest import check_is_fitted, _partition_estimators, Parallel, delayed, _accumulate_prediction
from molalkit.models.base import BaseSklearnModel


class RFRegressor(RandomForestRegressor, BaseSklearnModel):
    def fit_alb(self, train_data):
        return self.fit_alb_(train_data, self)

    def predict_uncertainty(self, pred_data):
        X = pred_data.X
        check_is_fitted(self)
        # Check data
        X = self._validate_X_predict(X)
        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        results = Parallel(
            n_jobs=n_jobs, verbose=self.verbose, prefer='processes')(
            delayed(e.predict)(X)
            for e in self.estimators_)
        return np.asarray(results).std(axis=0)

    def predict_value(self, pred_data):
        X = pred_data.X
        return super().predict(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os


def create_logger(name: str, save_dir: str = None, quiet: bool = False) -> logging.Logger:
    if name in logging.root.manager.loggerDict:
        return logging.getLogger(name)

    logger = logging.getLogger(name)

    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # Set logger depending on desired verbosity
    ch = logging.StreamHandler()
    if quiet:
        ch.setLevel(logging.INFO)
    else:
        ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)

        fh_v = logging.FileHandler(os.path.join(save_dir, '%s_verbose.log' % name))
        fh_v.setLevel(logging.DEBUG)
        # fh_q = logging.FileHandler(os.path.join(save_dir, '%s_quiet.log' % name))
        # fh_q.setLevel(logging.INFO)

        logger.addHandler(fh_v)
        # logger.addHandler(fh_q)

    return logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple, Callable
import os
import json
import pickle
import shutil
import pandas as pd
import numpy as np
from logging import Logger
from sklearn.metrics import *
from ..args import Metric
from .selection_method import BaseSelectionMethod, RandomSelectionMethod
from .forgetter import BaseForgetter, RandomForgetter, FirstForgetter


def eval_metric_func(y, y_pred, metric: str) -> float:
    if metric == 'roc-auc':
        return roc_auc_score(y, y_pred)
    elif metric == 'accuracy':
        y_pred = [1 if i >= 0.5 else 0 for i in y_pred]
        return accuracy_score(y, y_pred)
    elif metric == 'precision':
        y_pred = [1 if i >= 0.5 else 0 for i in y_pred]
        return precision_score(y, y_pred, average='macro')
    elif metric == 'recall':
        y_pred = [1 if i >= 0.5 else 0 for i in y_pred]
        return recall_score(y, y_pred, average='macro')
    elif metric == 'f1_score':
        y_pred = [1 if i >= 0.5 else 0 for i in y_pred]
        return f1_score(y, y_pred, average='macro')
    elif metric == 'mcc':
        y_pred = [1 if i >= 0.5 else 0 for i in y_pred]
        return matthews_corrcoef(y, y_pred)
    elif metric == 'r2':
        return r2_score(y, y_pred)
    elif metric == 'mae':
        return mean_absolute_error(y, y_pred)
    elif metric == 'mse':
        return mean_squared_error(y, y_pred)
    elif metric == 'rmse':
        return np.sqrt(eval_metric_func(y, y_pred, 'mse'))
    elif metric == 'max':
        return np.max(abs(y - y_pred))
    else:
        raise RuntimeError(f'Unsupported metrics {metric}')


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


class ActiveLearningResult:
    def __init__(self, n_iter: int):
        self.n_iter = n_iter
        self.results = dict()


class ActiveLearningTrajectory:
    def __init__(self, metrics: List[str] = None, top_k_score: bool = False):
        self.metrics = metrics
        self.top_k_score = top_k_score
        self.results = []

    def is_valid_column(self, column_name: str) -> Tuple[bool, List]:
        results = [alr.results.get(column_name) for alr in self.results]
        for result in results:
            if isinstance(result, float):
                return True, results
        return False, results

    def get_results(self) -> Dict:
        results_dict = dict()
        results_dict['n_iter'] = [alr.n_iter for alr in self.results]
        if self.metrics is not None:
            for metric in self.metrics:
                is_valid, results = self.is_valid_column(f'{metric}_selector')
                if is_valid:
                    results_dict[f'{metric}_selector'] = results
                for i in range(100):
                    is_valid, results = self.is_valid_column(f'{metric}_evaluator_{i}')
                    if is_valid:
                        results_dict[f'{metric}_evaluator_{i}'] = results
                    else:
                        break
        if self.top_k_score:
            is_valid, results = self.is_valid_column('top_k_score')
            if is_valid:
                results_dict['top_k_score'] = results
        results_dict['id_add'] = [json.dumps(alr.id_add, cls=NpEncoder) for alr in self.results]
        results_dict['acquisition_add'] = [json.dumps(alr.acquisition_add, cls=NpEncoder) for alr in self.results]
        results_dict['id_forgotten'] = [json.dumps(alr.id_forgotten, cls=NpEncoder) for alr in self.results]
        results_dict['acquisition_forgotten'] = [json.dumps(alr.acquisition_forgotten,
                                                            cls=NpEncoder) for alr in self.results]
        results_dict['id_prior_al'] = [json.dumps(alr.id_prior_al, cls=NpEncoder) for alr in self.results]
        return results_dict


class ActiveLearner:
    def __init__(self, save_dir: str, selection_method: BaseSelectionMethod, forgetter: BaseForgetter,
                 model_selector, dataset_train_selector, dataset_pool_selector, dataset_val_selector,
                 metrics: List[Metric], top_k_id: List[int] = None,
                 model_evaluators=None, dataset_train_evaluators=None,
                 dataset_pool_evaluators=None, dataset_val_evaluators=None,
                 yoked_learning_only: bool = False,
                 stop_size: int = None, evaluate_stride: int = None, kernel: Callable = None,
                 save_cpt_stride: int = None,
                 seed: int = 0,
                 logger: Logger = None):
        self.save_dir = save_dir
        self.selection_method = selection_method
        self.forgetter = forgetter
        self.model_selector = model_selector
        self.dataset_train_selector = dataset_train_selector
        self.dataset_pool_selector = dataset_pool_selector
        self.dataset_val_selector = dataset_val_selector
        self.metrics = metrics
        self.top_k_id = top_k_id
        self.model_evaluators = model_evaluators or []
        self.dataset_train_evaluators = dataset_train_evaluators or []
        self.dataset_pool_evaluators = dataset_pool_evaluators or []
        self.dataset_val_evaluators = dataset_val_evaluators or []
        self.yoked_learning_only = yoked_learning_only

        self.stop_size = stop_size
        self.evaluate_stride = evaluate_stride
        self.kernel = kernel  # used for cluster selection method
        self.save_cpt_stride = save_cpt_stride

        self.seed = seed
        if logger is not None:
            self.info = logger.info
        else:
            self.info = print
        self.current_iter = 0
        self.active_learning_traj = ActiveLearningTrajectory(metrics=self.metrics,
                                                             top_k_score=self.top_k_id is not None)

    @property
    def train_size(self) -> int:
        return len(self.dataset_train_selector)

    @property
    def pool_size(self) -> int:
        return len(self.dataset_pool_selector)

    def termination(self) -> bool:
        if self.pool_size == 0:
            self.info('Terminating active learning: pool size = 0')
            return True
        elif self.stop_size is not None and self.train_size >= self.stop_size:
            self.info(f'Terminating active learning: train_size > stop_size {self.stop_size}')
            return True
        else:
            return False

    def run(self, max_iter: int = None):
        info = f'Start active learning with training set size = {self.train_size}, pool set size = {self.pool_size}.' \
               f'\nSelection method: {self.selection_method.info}.'
        if self.forgetter is not None:
            info += f'\nForgetter: {self.forgetter.info}.'
        self.info(info)
        self.model_fitted = False
        for n_iter in range(self.current_iter, max_iter):
            alr = ActiveLearningResult(n_iter)
            alr.id_prior_al = self.get_id(self.dataset_train_selector)
            if self.termination():
                break
            self.info('Start an new iteration of active learning: %d.' % self.current_iter)
            # evaluate
            if self.evaluate_stride is not None and n_iter % self.evaluate_stride == 0:
                self.evaluate(alr)
            # add sample
            self.add_samples(alr)
            # forget sample
            self.forget_samples(alr)
            # save checkpoint file
            if self.save_cpt_stride is not None and n_iter % self.save_cpt_stride == 0:
                self.save(path=self.save_dir, filename='al_temp.pkl', overwrite=True)
                shutil.move(os.path.join(self.save_dir, 'al_temp.pkl'), os.path.join(self.save_dir, 'al.pkl'))
                self.info('save checkpoint file %s/al.pkl' % self.save_dir)

            self.current_iter += 1
            self.info('Training set size = %i' % self.train_size)
            self.info('Pool set size = %i' % self.pool_size)
            self.active_learning_traj.results.append(alr)

        df_traj = pd.DataFrame(self.active_learning_traj.get_results())
        df_traj.to_csv(os.path.join(self.save_dir, 'al_traj.csv'), index=False)
        if self.save_cpt_stride:
            self.save(path=self.save_dir, overwrite=True)

    def evaluate(self, alr: ActiveLearningResult):
        self.info('evaluating model performance.')
        # conventional active learning.
        if not self.yoked_learning_only:
            # evaluate the prediction performance of ML model on the validation set
            if self.metrics is not None:
                if not self.model_fitted:
                    self.model_selector.fit_alb(self.dataset_train_selector)
                    self.model_fitted = True
                y_pred = self.model_selector.predict_value(self.dataset_val_selector)
                for metric in self.metrics:
                    metric_value = eval_metric_func(self.dataset_val_selector.y, y_pred, metric=metric)
                    alr.results[f'{metric}_selector'] = metric_value
            # evaluate the percentage of top-k data selected in the training set
            if self.top_k_id is not None:
                alr.results['top_k_score'] = self.get_top_k_score(self.dataset_train_selector, self.top_k_id)
        # yoked learning
        for i, model in enumerate(self.model_evaluators):
            if self.metrics is not None:
                model.fit_alb(self.dataset_train_evaluators[i])
                y_pred = model.predict_value(self.dataset_val_evaluators[i])
                for metric in self.metrics:
                    metric_value = eval_metric_func(self.dataset_val_evaluators[i].y, y_pred, metric=metric)
                    alr.results[f'{metric}_evaluator_{i}'] = metric_value
        self.info('evaluating model performance finished.')

    def add_samples(self, alr: ActiveLearningResult):
        # train the model if it is not trained in the evaluation step, and the selection method is not random.
        if not self.model_fitted and not isinstance(self.selection_method, RandomSelectionMethod):
            self.model_selector.fit_alb(self.dataset_train_selector)
        selected_idx, acquisition = self.selection_method(model=self.model_selector,
                                                          data_train=self.dataset_train_selector,
                                                          data_pool=self.dataset_pool_selector,
                                                          kernel=self.kernel)
        alr.id_add = [self.dataset_pool_selector.data[i].id for i in selected_idx]
        alr.acquisition_add = acquisition
        # transfer data from pool to train.
        for i in sorted(selected_idx, reverse=True):
            self.dataset_train_selector.data.append(self.dataset_pool_selector.data.pop(i))
            for j in range(len(self.model_evaluators)):
                self.dataset_train_evaluators[j].data.append(self.dataset_pool_evaluators[j].data.pop(i))
        # set the model unfitted because new data is added.
        self.model_fitted = False

    def forget_samples(self, alr: ActiveLearningResult):
        # no forget algorithm is applied.
        # or forget algorithm is applied when self.forgetter.forget_size <= self.train_size.
        if self.forgetter is None or (self.forgetter.forget_size is not None and
                                      self.forgetter.forget_size > self.train_size):
            alr.id_forgotten = []
            alr.acquisition_forgotten = []
            return
        # train the model if the forgetter is not random or first.
        if not self.model_fitted and not self.forgetter.__class__ in [RandomForgetter, FirstForgetter]:
            self.model_selector.fit_alb(self.dataset_train_selector)
        # forget algorithm is applied.
        forgotten_idx, acquisition = self.forgetter(model=self.model_selector,
                                                    data=self.dataset_train_selector,
                                                    batch_size=self.forgetter.batch_size,
                                                    cutoff=self.forgetter.forget_cutoff)
        if forgotten_idx:
            alr.id_forgotten = [self.dataset_train_selector.data[i].id for i in forgotten_idx]
            alr.acquisition_forgotten = acquisition
            # transfer data from train to pool.
            for i in sorted(forgotten_idx, reverse=True):
                self.dataset_pool_selector.data.append(self.dataset_train_selector.data.pop(i))
                for j in range(len(self.model_evaluators)):
                    self.dataset_pool_evaluators[j].data.append(self.dataset_train_evaluators[j].data.pop(i))
        else:
            alr.id_forgotten = []
            alr.acquisition_forgotten = []

    @staticmethod
    def get_id(dataset):
        id_list = []
        for data in dataset:
            id_list.append(data.id)
        return id_list

    @staticmethod
    def get_top_k_score(dataset, top_k_id) -> float:
        N_top_k = 0
        for data in dataset:
            if data.id in top_k_id:
                N_top_k += 1
        return N_top_k / len(top_k_id)

    def save(self, path, filename='al.pkl', overwrite=False):
        f_al = os.path.join(path, filename)
        if os.path.isfile(f_al) and not overwrite:
            raise RuntimeError(
                f'Path {f_al} already exists. To overwrite, set '
                '`overwrite=True`.'
            )
        store = self.__dict__.copy()
        pickle.dump(store, open(f_al, 'wb'), protocol=4)

    @classmethod
    def load(cls, path, filename='al.pkl'):
        f_al = os.path.join(path, filename)
        store = pickle.load(open(f_al, 'rb'))
        input = {}
        for key in ['save_dir', 'dataset_type', 'metrics', 'learning_type', 'model_selector', 'dataset_train_selector',
                    'dataset_pool_selector', 'dataset_val_evaluator', 'df_train', 'df_pool']:
            input[key] = store[key]
        dataset = cls(**input)
        dataset.__dict__.update(**store)
        return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple
import copy
import os
import json
import pickle
import numpy as np
import pandas as pd
import rdkit.Chem.AllChem as Chem
from joblib import Parallel, delayed
from sklearn.preprocessing import StandardScaler
from mgktools.features_mol import FeaturesGenerator
from mgktools.data.data import tolist, to_numpy
# Cache of RDKit molecules
CACHE_MOL = True
SMILES_TO_MOL: Dict[str, Chem.Mol] = {}


def remove_none(X: List):
    X_ = []
    for x in X:
        if x is not None and len(x) != 0:
            X_.append(x)
    if len(X_) == 0:
        return None
    else:
        return X_


def concatenate(X: List, axis: int = 0, dtype=None):
    X_ = remove_none(X)
    if X_:
        return np.concatenate(X_, axis=axis, dtype=dtype)
    else:
        return None


class DataPointPure:
    """
    This is the class that stores a data point containing one molecule. (pure substance)

    Parameters
    ----------
    smiles: str
        SMILES string of the molecule
    mol:
        molecule object in RDKit
    fingerprints:
        1-d numpy array
    """

    def __init__(self, smiles: str):
        self.smiles = smiles
        self.fingerprints = None

    def __repr__(self) -> str:
        return self.smiles

    @property
    def mol(self):
        if self.smiles in SMILES_TO_MOL:
            return SMILES_TO_MOL[self.smiles]
        else:
            mol = Chem.MolFromSmiles(self.smiles)
            SMILES_TO_MOL[self.smiles] = mol
            return mol

    @property
    def mols(self):
        return [self.mol]

    def set_fingerprints(self, features_generator: List[FeaturesGenerator]):
        fingerprints = []
        for fg in features_generator:
            fingerprints.append(self.calc_features_mol(self.mol, fg))
        fingerprints = np.concatenate(fingerprints)
        # Fix nans in features_mol
        replace_token = 0
        self.fingerprints = np.where(np.isnan(fingerprints), replace_token, fingerprints)

    @staticmethod
    def calc_features_mol(mol: Chem.Mol, features_generator: FeaturesGenerator):
        if mol is not None and mol.GetNumHeavyAtoms() > 0:
            features_mol = features_generator(mol)
        elif mol is not None and mol.GetNumHeavyAtoms() == 0:
            # Hydrogen
            # not all features_mol are equally long, so use methane as dummy molecule to determine length
            features_mol = np.zeros(len(features_generator(Chem.MolFromSmiles('C'))))
        else:
            raise ValueError('rdkit mol object cannot be None.')
        return np.asarray(features_mol)


class DataPointMix:
    """
    This is the class that stores a data point containing more than one molecule. (mixture substances)

    Parameters
    ----------
    data:
        list of DatapointPure object
    concentration: list of float
        the mole concentration of molecules
    """

    def __init__(self, data: List[DataPointPure],
                 concentration: List[float] = None):
        assert len(data) == len(concentration)
        assert len(data) > 1
        # read data point
        self.data = data
        # set concentration
        if concentration is None:
            self.concentration = [1.0 / len(data)] * len(data)
        else:
            self.concentration = concentration
        self.fingerprints = None

    def __repr__(self) -> str:
        repr = ';'.join(list(map(lambda x, y: x.__repr__() + ',%.3f' % y, self.data, self.concentration)))
        return repr

    @property
    def smiles(self) -> List[str]:
        return [d.smiles for d in self.data]

    @property
    def mols(self) -> List[Chem.Mol]:
        return [d.mol for d in self.data]

    @classmethod
    def from_smiles_list(cls,
                         smiles: List[str],
                         concentration: List[float] = None):
        return cls([DataPointPure(s) for s in smiles], concentration=concentration)

    def set_fingerprints(self, features_generator: List[FeaturesGenerator],
                         features_combination: Literal['concat', 'mean'] = None):
        if features_generator is None:
            return
        if len(self.data) != 1:
            self.fingerprints = []
            for i, d in enumerate(self.data):
                d.set_fingerprints(features_generator)
                self.fingerprints.append(d.fingerprints * self.concentration[i])
            if features_combination == 'mean':
                self.fingerprints = np.mean(self.fingerprints, axis=0)
            elif features_combination == 'concat':
                self.fingerprints = np.concatenate(self.fingerprints)
            else:
                raise ValueError(f'unknown feature combination: f{features_combination}')
        else:
            self.data[0].set_fingerprints(features_generator)
            self.fingerprints = self.data[0].fingerprints


class DataPoint:
    """
    This is the class that stores a data point containing more than one molecule. (mixture substances)

    Parameters
    ----------
    data:
        list of DatapointPure object
    targets: list of float
        target properties to be predicted
    concentration: list of float
        the mole concentration of molecules
    features_add: list of float
        additional features, such as temperature, pressure, etc.
    """
    def __init__(self,
                 data_p: List[DataPointPure],
                 data_m: List[DataPointMix],
                 targets: List[float],
                 features_add: List[float] = None):
        # read data point
        self.data_p = data_p or []
        self.data_m = data_m or []
        self.data = self.data_p + self.data_m
        self.targets = targets
        self.features_add = features_add
        self.fingerprints = None

    def __repr__(self) -> str:
        repr = ';'.join(list(map(lambda x : x.__repr__(), self.data)))
        if self.features_add is not None:
            repr += ';' + ','.join(list(map(lambda x: '%.5f' % x, self.features_add)))
        return repr

    @property
    def mols(self) -> List[List[Chem.Mol]]:
        return [d.mols for d in self.data]

    def set_fingerprints(self, features_generator: List[FeaturesGenerator],
                         features_combination: Literal['concat', 'mean'] = None):
        for d in self.data_p:
            d.set_fingerprints(features_generator)
        for d in self.data_m:
            d.set_fingerprints(features_generator, features_combination=features_combination)
        self.fingerprints = np.concatenate([d.fingerprints for d in self.data], axis=0)  # 1-d array

    @property
    def y(self) -> np.ndarray:  # List[List[float]]
        return np.array(self.targets, dtype=float)


class Dataset:
    def __init__(self, data: List[DataPoint] = None,
                 mgk_type: Literal['graph', 'pre-computed'] = None):
        self.data = data
        self.mgk_type = mgk_type
        self.fingerprints_scaler = None
        self.features_add_scaler = None

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, item) -> Union[DataPoint, List[DataPoint]]:
        return self.data[item]

    @property
    def fingerprints_raw(self) -> Optional[np.ndarray]:  # List[List[float]]
        return np.asarray([d.fingerprints for d in self.data])

    @property
    def fingerprints(self) -> np.ndarray:  # List[List[float]]
        fingerprints = self.fingerprints_raw
        if self.fingerprints_scaler is not None:
            fingerprints = self.fingerprints_scaler.transform(fingerprints)
        return fingerprints

    @property
    def features_add_raw(self) -> Optional[np.ndarray]:  # List[List[float]]
        if self.data[0].features_add is None:
            return None
        else:
            return np.asarray([d.features_add for d in self.data])

    @property
    def features_add(self) -> np.ndarray:  # List[List[float]]
        features_add = self.features_add_raw
        if self.features_add_scaler is not None:
            features_add = self.features_add_scaler.transform(features_add)
        return features_add

    @property
    def repr(self) -> np.ndarray:  # # List[List[str]]
        return np.asarray([d.__repr__() for d in self.data])

    @property
    def X(self) -> np.ndarray:
        return concatenate([self.fingerprints, self.features_add], axis=1)

    @property
    def y(self) -> np.ndarray:  # List[List[float]]
        y = concatenate([d.targets for d in self.data], axis=0)
        if y is not None and y.shape[1] == 1:
            return y.ravel()
        else:
            return y

    @property
    def N_tasks(self) -> int:
        return len(self.data[0].targets)

    @property
    def N_fingerprints(self) -> int:
        if self.fingerprints_raw is None:
            return 0
        else:
            return self.fingerprints_raw.shape[1]

    @property
    def N_features_add(self) -> int:
        if self.features_add_raw is None:
            return 0
        else:
            return self.features_add_raw.shape[1]

    def copy(self):
        return copy.deepcopy(self)

    def normalize_features(self, normalize_fingerprints: bool = False,
                           normalize_features_add: bool = False):
        if normalize_fingerprints:
            if self.fingerprints_raw is None:
                raise ValueError('You need to set_fingerprints before normalization.')
            self.fingerprints_scaler = StandardScaler().fit(self.fingerprints_raw)
        if normalize_features_add:
            if self.features_add_raw is None:
                raise ValueError('You need to set_fingerprints before normalization.')
            self.features_add_scaler = StandardScaler().fit(self.features_add_raw)

    def set_fingerprints(self, features_generator: List[FeaturesGenerator]):
        for d in self.data:
            d.set_fingerprints(features_generator)

    def save(self, path, filename='data.pkl', overwrite=False):
        f_dataset = os.path.join(path, filename)
        if os.path.isfile(f_dataset) and not overwrite:
            raise RuntimeError(
                f'Path {f_dataset} already exists. To overwrite, set '
                '`overwrite=True`.'
            )
        store = self.__dict__.copy()
        pickle.dump(store, open(f_dataset, 'wb'), protocol=4)

    @classmethod
    def load(cls, path, filename='data.pkl'):
        f_dataset = os.path.join(path, filename)
        store = pickle.load(open(f_dataset, 'rb'))
        dataset = cls()
        dataset.__dict__.update(**store)
        return dataset

    def features_size(self) -> int:
        return self.N_fingerprints + self.N_features_add

    @staticmethod
    def read_DataPoint(
            smiles: List[str],
            mixture: List[Union[str, float]],
            targets: List[float],
            features_add: List[float] = None,
            features_generator: List[FeaturesGenerator] = None,
            features_combination: Literal['concat', 'mean'] = None,
    ) -> DataPoint:
        data_p = [DataPointPure(s) for s in smiles] if smiles is not None else None
        data_m = [DataPointMix.from_smiles_list(m[0::2], m[1::2]) for m in mixture] if mixture is not None else None
        data = DataPoint(data_p=data_p,
                         data_m=data_m,
                         targets=targets,
                         features_add=features_add)
        data.set_fingerprints(features_generator, features_combination=features_combination)
        return data

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame,
                       pure_columns: List[str] = None,
                       mixture_columns: List[str] = None,
                       target_columns: List[str] = None,
                       feature_columns: List[str] = None,
                       features_generator: List[FeaturesGenerator] = None,
                       features_combination: Literal['concat', 'mean'] = None,
                       n_jobs: int = 8):
        data = Parallel(
            n_jobs=n_jobs, verbose=True, prefer='processes')(
            delayed(cls.read_DataPoint)(
                tolist(df.iloc[i].get(pure_columns)),
                tolist(df.iloc[i].get(mixture_columns)),
                to_numpy(df.iloc[i:i + 1][target_columns]),
                features_add=to_numpy(df.iloc[i:i + 1].get(feature_columns)),
                features_generator=features_generator,
                features_combination=features_combination,
            )
            for i in df.index)
        return cls(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List, Tuple
import numpy as np
from molalkit.models.random_forest.RandomForestClassifier import RFClassifier
from molalkit.models.gaussian_process.GaussianProcessRegressor import GPRegressor
from molalkit.al.selection_method import get_topn_idx


class BaseForgetter(ABC):
    def __init__(self, batch_size: int = 1, forget_size: int = 1, forget_cutoff: float = None):
        self.batch_size = batch_size
        self.forget_size = forget_size
        self.forget_cutoff = forget_cutoff

    @abstractmethod
    def __call__(self, **kwargs) -> Tuple[List[int], List[float]]:
        pass

    @property
    @abstractmethod
    def info(self) -> str:
        pass


class BaseRandomForgetter(BaseForgetter, ABC):
    """Base Forgetter that uses random seed."""
    def __init__(self, batch_size: int = 1, forget_size: int = 1, forget_cutoff: float = None, seed: int = 0):
        super().__init__(batch_size=batch_size, forget_size=forget_size, forget_cutoff=forget_cutoff)
        np.random.seed(seed)


class RandomForgetter(BaseRandomForgetter):
    def __call__(self, data, batch_size: int = 1, **kwargs) -> Tuple[List[int], None]:
        assert batch_size < len(data)
        return np.random.choice(list(range(len(data))), batch_size, replace=False).tolist(), None

    @property
    def info(self) -> str:
        return 'RandomForgetter'


class FirstForgetter(BaseForgetter):
    def __call__(self, data, batch_size: int = 1, **kwargs) -> Tuple[List[int], None]:
        assert batch_size < len(data)
        return list(range(batch_size)), None

    @property
    def info(self) -> str:
        return 'FirstForgetter'


class MinOOBUncertaintyForgetter(BaseRandomForgetter):
    def __call__(self, model: RFClassifier, data, batch_size: int = 1, **kwargs) -> Tuple[List[int], List[float]]:
        assert batch_size < len(data)
        assert isinstance(model, RFClassifier)
        assert model.oob_score is True
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        y_oob_uncertainty = (0.25 - np.var(y_oob_proba, axis=1)) * 4
        # select the top-n points with least uncertainty
        forgotten_idx = get_topn_idx(y_oob_uncertainty, n=batch_size, target='min')
        acquisition = y_oob_uncertainty[np.array(forgotten_idx)].tolist()
        return forgotten_idx, acquisition

    @property
    def info(self) -> str:
        return 'MinOOBUncertaintyForgetter'


class MaxOOBUncertaintyForgetter(BaseRandomForgetter):
    def __call__(self, model: RFClassifier, data, batch_size: int = 1, **kwargs) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the highest out-of-bag (OOB) uncertainty.

        Parameters
        ----------
        model: Only random forest classifier is supported due to efficient OOB uncertainty calculation.
        data: The dataset to forget.
        batch_size: The number of samples to forget.

        Returns
        -------
        The index of samples to forget.
        """
        assert batch_size < len(data)
        assert isinstance(model, RFClassifier)
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        y_oob_uncertainty = (0.25 - np.var(y_oob_proba, axis=1)) * 4
        # select the top-n points with least uncertainty
        forgotten_idx = get_topn_idx(y_oob_uncertainty, n=batch_size)
        acquisition = y_oob_uncertainty[np.array(forgotten_idx)].tolist()
        return forgotten_idx, acquisition

    @property
    def info(self) -> str:
        return 'MaxOOBUncertaintyForgetter'


class MinOOBErrorForgetter(BaseRandomForgetter):
    def __call__(self, model: RFClassifier, data, batch_size: int = 1, cutoff: float = None, **kwargs
                 ) -> Tuple[List[int], List[float]]:
        assert batch_size < len(data)
        assert isinstance(model, RFClassifier)
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        oob_error = np.absolute(y_oob_proba[:, 1] - data.y)
        # select the top-n points with least uncertainty
        forgotten_idx = get_topn_idx(oob_error, n=batch_size, target='min', cutoff=cutoff)
        acquisition = oob_error[np.array(forgotten_idx)].tolist() if forgotten_idx else []
        return forgotten_idx, acquisition

    @property
    def info(self) -> str:
        return 'MinOOBErrorForgetter'


class MinLOOErrorForgetter(BaseRandomForgetter):
    def __call__(self, model: GPRegressor, data, batch_size: int = 1, cutoff: float = None, **kwargs
                 ) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the lowest Leave-one-out cross-validation (LOOCV) error.
        Parameters
        ----------
        model: Only Gaussian process regressor is supported due to efficient LOOCV of GPR.
        data: The dataset to forget.
        batch_size: The number of samples to forget.
        cutoff: The cutoff value of LOOCV error. Only samples with LOOCV error lower than cutoff will be forgot.

        Returns
        -------
        The index and the acquisition value of samples to forget.
        """
        assert batch_size < len(data)
        assert isinstance(model, GPRegressor)
        y_loocv = model.predict_loocv(data.X, data.y, return_std=False)
        # uncertainty calculation, normalized into 0 to 1
        loo_error = np.absolute(y_loocv - data.y)
        # select the top-n points with least uncertainty
        forgotten_idx = get_topn_idx(loo_error, n=batch_size, target='min', cutoff=cutoff)
        acquisition = loo_error[np.array(forgotten_idx)].tolist() if forgotten_idx else []
        return forgotten_idx, acquisition

    @property
    def info(self) -> str:
        return 'MinLOOErrorForgetter'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import math
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple
import copy
import os
import pickle
import numpy as np
import pandas as pd
from random import Random
import rdkit.Chem.AllChem as Chem
from chemprop.data.scaffold import scaffold_to_smiles
from logging import Logger
from .data import Dataset


def get_data(path: str,
             pure_columns: List[str] = None,
             mixture_columns: List[str] = None,
             target_columns: List[str] = None,
             feature_columns: List[str] = None,
             features_generator: List[str] = None,
             features_combination: Literal['concat', 'mean'] = None,
             n_jobs: int = 8):
    df = pd.read_csv(path)
    return Dataset.from_dataframe(df,
                                  pure_columns=pure_columns,
                                  mixture_columns=mixture_columns,
                                  target_columns=target_columns,
                                  feature_columns=feature_columns,
                                  features_generator=features_generator,
                                  features_combination=features_combination,
                                  n_jobs=n_jobs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import os
import shutil
from molalkit.args import ActiveLearningArgs
from test_model.test_model import run, al_results_check


CWD = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize('learning_type', ['explorative', 'passive'])
def test_classification(learning_type):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'carcinogens_lagunin',
        '--metrics', 'roc-auc',
        '--learning_type', learning_type,
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'scaffold_order',
        '--split_sizes', '0.5', '0.5',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


@pytest.mark.parametrize('learning_type', ['explorative', 'passive'])
def test_regression(learning_type):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'test_regression',
        '--metrics', 'rmse',
        '--learning_type', learning_type,
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'scaffold_order',
        '--split_sizes', '0.5', '0.5',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


@pytest.mark.parametrize('learning_type', ['exploitive'])
@pytest.mark.parametrize('exploitive_target', ['min', 'max', '1.0'])
def test_regression_exploitive(learning_type, exploitive_target):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'test_regression',
        '--metrics', 'rmse',
        '--learning_type', learning_type,
        '--exploitive_target', exploitive_target,
        '--top_k', '0.1',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'scaffold_order',
        '--split_sizes', '0.5', '0.5',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple
from functools import cached_property
from tap import Tap
import os
import shutil
import json
import math
import pandas as pd
import numpy as np
from mgktools.features_mol import FeaturesGenerator
from mgktools.data.split import data_split_index
from molalkit.logging import create_logger
from molalkit.utils import get_data, get_model, get_kernel
from molalkit.data.datasets import DATA_DIR
from molalkit.al.selection_method import *
from molalkit.al.forgetter import *

CWD = os.path.dirname(os.path.abspath(__file__))
Metric = Literal['roc-auc', 'accuracy', 'precision', 'recall', 'f1_score', 'mcc',
                 'rmse', 'mae', 'mse', 'r2', 'max']


class CommonArgs(Tap):
    save_dir: str
    """the output directory."""
    n_jobs: int = 1
    """the cpu numbers used for parallel computing."""
    quiet: bool = False
    """Whether the stream handler should be quiet (i.e., print only important info)."""
    logger_name: str = 'alb_output'
    """the prefix of the output logger file: verbose.log and quite.log"""
    seed: int = 0
    """random seed."""

    def process_args(self) -> None:
        os.makedirs(self.save_dir, exist_ok=True)
        self.logger = create_logger(self.logger_name, save_dir=self.save_dir, quiet=self.quiet)
        np.random.seed(self.seed)


class DatasetArgs(CommonArgs):
    data_public = None
    """Use public data sets."""
    data_path: str = None
    """the Path of input data CSV file."""
    data_path_training: str = None
    """the Path of input data CSV file for training set."""
    data_path_pool: str = None
    """the Path of input data CSV file for pool set."""
    data_path_val: str = None
    """the Path of input data CSV file for validation set."""
    pure_columns: List[str] = None
    """
    For pure compounds.
    Name of the columns containing single SMILES or InChI string.
    """
    mixture_columns: List[str] = None
    """
    For mixtures.
    Name of the columns containing multiple SMILES or InChI string and 
    corresponding concentration.
    example: ['C', 0.5, 'CC', 0.3]
    """
    target_columns: List[str] = None
    """
    Name of the columns containing target values.
    """
    feature_columns: List[str] = None
    """
    Name of the columns containing additional features_mol such as temperature, 
    pressuer.
    """
    dataset_type: Literal['regression', 'classification', 'multiclass'] = None
    """
    Type of task.
    """
    split_type: Literal['random', 'scaffold_random', 'scaffold_order'] = None
    """Method of splitting the data into active learning/validation."""
    split_sizes: List[float] = None
    """Split proportions for active learning/validation sets."""
    full_val: bool = False
    """validate the performance of active learning on the full dataset."""
    error_rate: float = None
    """the percent of the training set that will be affected by error (0-1)."""

    def process_args(self) -> None:
        super().process_args()
        if self.data_public == 'freesolv' or self.data_public == 'test_regression':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['freesolv']
            self.dataset_type = 'regression'
        elif self.data_public == 'delaney':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['logSolubility']
            self.dataset_type = 'regression'
        elif self.data_public == 'lipo':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['lipo']
            self.dataset_type = 'regression'
        elif self.data_public == 'pdbbind_refined':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['-logKd/Ki']
            self.dataset_type = 'regression'
        elif self.data_public == 'pdbbind_full':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['-logKd/Ki']
            self.dataset_type = 'regression'
        elif self.data_public in ['ld50_zhu', 'caco2_wang', 'solubility_aqsoldb', 'ppbr_az', 'vdss_lombardo',
                                  'Half_Life_Obach', 'Clearance_Hepatocyte_AZ']:
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['Drug']
            self.target_columns = ['Y']
            self.dataset_type = 'regression'
        elif self.data_public == 'bbbp':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['p_np']
            self.dataset_type = 'classification'
        elif self.data_public == 'bace':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['mol']
            self.target_columns = ['Class']
            self.dataset_type = 'classification'
        elif self.data_public == 'hiv':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['HIV_active']
            self.dataset_type = 'classification'
        elif self.data_public in ['ames', 'carcinogens_lagunin', 'dili', 'herg', 'skin', 'hia_hou', 'pgp_broccatelli',
                                  'bioavailability_ma', 'clintox', 'bbb_martins', 'CYP1A2_Veith',
                                  'CYP2C9_Substrate_CarbonMangels', 'CYP2C9_Veith', 'CYP2C19_Veith',
                                  'CYP2D6_Substrate_CarbonMangels', 'CYP2D6_Veith', 'CYP3A4_Veith',
                                  'CYP3A4_Substrate_CarbonMangels']:
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['Drug']
            self.target_columns = ['Y']
            self.dataset_type = 'classification'


        if self.split_type == 'scaffold':
            assert len(self.pure_columns) == 1
            assert self.mixture_columns is None

        assert len(self.target_columns) == 1

        if self.data_path is not None:
            # All data comes from the same file.
            assert self.data_path_val is None and self.data_path_training is None and self.data_path_pool is None
            df = pd.read_csv(self.data_path)
            if 'id' not in df:
                df['id'] = range(len(df))
            if self.full_val:
                # use the full dataset as validation set
                assert self.split_type is None
                assert self.split_sizes is None
                assert self.error_rate is None
                df.to_csv('%s/val.csv' % self.save_dir, index=False)
                df_al = df
            else:
                # split the dataset into active learning and validation sets
                al_index, val_index = data_split_index(
                    n_samples=len(df),
                    mols=df[self.pure_columns[0]] if self.pure_columns is not None else None,
                    # targets=df[self.target_columns[0]],
                    split_type=self.split_type,
                    sizes=self.split_sizes,
                    seed=self.seed,
                    logger=self.logger)
                df[df.index.isin(val_index)].to_csv('%s/val.csv' % self.save_dir, index=False)
                df_al = df[df.index.isin(al_index)]
                if self.error_rate is not None:
                    # randomly select a portion of the training set to be affected by error
                    error_index = np.random.choice(al_index, int(self.error_rate * len(al_index)), replace=False)
                    df_al.loc[df_al.index.isin(error_index), self.target_columns[0]] ^= 1
                    df_al.loc[df_al.index.isin(error_index), 'flip_label'] = True
                    df_al.loc[~df_al.index.isin(error_index), 'flip_label'] = False
            # split the active learning set into training and pool sets
            assert self.init_size < len(df_al)
            if self.dataset_type == 'regression':
                train_index, pool_index = data_split_index(
                    n_samples=len(df_al),
                    mols=df_al[self.pure_columns[0]] if self.pure_columns is not None else None,
                    split_type='random',
                    sizes=[self.init_size / len(df_al), 1 - self.init_size / len(df_al)],
                    seed=self.seed)
            else:
                train_index, pool_index = data_split_index(
                    n_samples=len(df_al),
                    mols=df_al[self.pure_columns[0]] if self.pure_columns is not None else None,
                    targets=df_al[self.target_columns[0]],
                    split_type='init_al',
                    n_samples_per_class=1,
                    seed=self.seed)
                # randomly select self.init_size - 2 samples from the pool set to be the training set
                if self.init_size > 2:
                    train_index.extend(np.random.choice(pool_index, self.init_size - 2, replace=False))
                    _ = []
                    for i in pool_index:
                        if i not in train_index:
                            _.append(i)
                    pool_index = _
            df_al.iloc[train_index].to_csv('%s/train_init.csv' % self.save_dir, index=False)
            df_al.iloc[pool_index].to_csv('%s/pool_init.csv' % self.save_dir, index=False)
        else:
            # data comes from 3 different files.
            assert self.data_path_training is not None, 'please provide input data'
            assert self.data_path_pool is not None, 'please provide input data'
            assert self.data_path_val is not None, 'please provide input data'
            shutil.copyfile(self.data_path_training, '%s/train_init.csv' % self.save_dir)
            shutil.copyfile(self.data_path_pool, '%s/pool_init.csv' % self.save_dir)
            shutil.copyfile(self.data_path_val, '%s/val.csv' % self.save_dir)
            df = pd.concat([pd.read_csv(f) for f in [self.data_path_training, self.data_path_pool, self.data_path_val]])
            if 'id' not in df:
                df['id'] = range(len(df))
                df_train = pd.read_csv('%s/train_init.csv' % self.save_dir)
                df_train['id'] = range(len(df_train))
                df_train.to_csv('%s/train_init.csv' % self.save_dir)
                df_pool = pd.read_csv('%s/pool_init.csv' % self.save_dir)
                df_pool['id'] = range(len(df_pool))
                df_pool['id'] += len(df_train)
                df_pool.to_csv('%s/pool_init.csv' % self.save_dir)
                df_val = pd.read_csv('%s/val.csv' % self.save_dir)
                df_val['id'] = range(len(df_val))
                df_val['id'] += len(df_train) + len(df_pool)
                df_val.to_csv('%s/val.csv' % self.save_dir)
            df.to_csv('%s/full.csv' % self.save_dir)
            self.data_path = '%s/full.csv' % self.save_dir


class ModelArgs(Tap):
    model_config_selector: str
    """config file contain all information of the machine learning model."""
    model_config_evaluators: List[str] = None
    """A list of config files contain all information of the machine learning model for performance evaluation."""

    @property
    def model_config_selector_dict(self) -> Dict:
        config_name = self.find_stored_model_config(self.model_config_selector)
        return json.loads(open(config_name).read())

    @property
    def model_config_evaluators_dict(self) -> List[Dict]:
        if self.model_config_evaluators is None:
            return []
        else:
            return [json.loads(open(self.find_stored_model_config(m)).read()) for m in self.model_config_evaluators]

    @staticmethod
    def find_stored_model_config(config_name):
        if os.path.exists(config_name):
            return config_name
        else:
            stored_config_name = f'{CWD}/models/configs/{config_name}'
            assert os.path.exists(stored_config_name), f'{config_name} is not a valid model configuration'
            return stored_config_name


class ActiveLearningArgs(DatasetArgs, ModelArgs):
    save_dir: str
    """the output directory."""
    n_jobs: int = 1
    """the cpu numbers used for parallel computing."""
    data_path: str = None
    """the Path of input data CSV file."""
    metrics: List[Metric]
    """the metrics to evaluate model performance."""
    evaluate_stride: int = 100
    """evaluate model performance on the validation set when the size of the training set is an integer multiple of the 
    evaluation stride."""
    top_k: float = None
    """the ratio of top molecules are considered."""
    yoked_learning_only: bool = False
    """Only perform yoked learning."""
    learning_type: Literal['passive', 'explorative', 'exploitive']
    """the learning type to be performed."""
    exploitive_target: str = None
    """the target value for exploitive active learning."""
    init_size: int = 2
    """number of samples as the initial."""
    batch_size: int = 1
    """number of samples added in each active learning iteration."""
    batch_mode: Literal['naive', 'clustering'] = 'naive'
    """the method that add a batch of samples."""
    stop_ratio: float = None
    """Stop active learning when the selected molecules reach the ratio."""
    stop_size: int = None
    """Stop active learning when the selected molecules reach the number."""
    max_iter: int = None
    """the maximum number of iterations."""
    save_cpt_stride: int = None
    """save checkpoint file every no. steps of active learning iteration."""
    load_checkpoint: bool = False
    """load checkpoint file and continue the active learning."""
    # Arguments for forgetting active learning.
    forget_protocol: Literal['forget_first', 'forget_random', 'min_oob_uncertainty', 'max_oob_uncertainty',
    'min_oob_error', 'min_loo_error'] = None
    """protocol to use (forget_first, forget_random, min_oob_uncertain (RF only), max_oob_uncertain (RF only)
    , min_loo_error)."""
    forget_cutoff: float = None
    """The error cutoff for forgetting ."""
    forget_size: int = None
    """the number of molecules in the training set to start forgetting data at."""
    forget_ratio: float = None
    """the percent of the full training set to start forgetting data."""

    @property
    def model_selector(self):
        if not hasattr(self, '_model_selector'):
            self._model_selector = get_model(
                data_format=self.model_config_selector_dict['data_format'],
                dataset_type=self.dataset_type,
                model=self.model_config_selector_dict.get('model'),
                save_dir='%s/selector' % self.save_dir,
                loss_function=self.model_config_selector_dict.get('loss_function'),
                num_tasks=len(self.target_columns),
                multiclass_num_classes=self.model_config_selector_dict.get('loss_function') or 3,
                features_generator=self.features_generator_selector,
                no_features_scaling=self.model_config_selector_dict.get('no_features_scaling') or False,
                features_only=self.model_config_selector_dict.get('features_only') or False,
                features_size=self.data_train_selector.features_size(),
                epochs=self.model_config_selector_dict.get('epochs') or 30,
                depth=self.model_config_selector_dict.get('depth') or 3,
                hidden_size=self.model_config_selector_dict.get('hidden_size') or 300,
                ffn_num_layers=self.model_config_selector_dict.get('ffn_num_layers') or 2,
                ffn_hidden_size=self.model_config_selector_dict.get('ffn_hidden_size'),
                dropout=self.model_config_selector_dict.get('dropout') or 0.0,
                batch_size=self.model_config_selector_dict.get('batch_size') or 50,
                ensemble_size=self.model_config_selector_dict.get('ensemble_size') or 1,
                number_of_molecules=self.model_config_selector_dict.get('number_of_molecules') or 1,
                mpn_shared=self.model_config_selector_dict.get('mpn_shared') or False,
                atom_messages=self.model_config_selector_dict.get('atom_messages') or False,
                undirected=self.model_config_selector_dict.get('undirected') or False,
                class_balance=self.model_config_selector_dict.get('class_balance') or False,
                checkpoint_dir=self.model_config_selector_dict.get('checkpoint_dir'),
                checkpoint_frzn=self.model_config_selector_dict.get('checkpoint_frzn'),
                frzn_ffn_layers=self.model_config_selector_dict.get('frzn_ffn_layers') or 0,
                freeze_first_only=self.model_config_selector_dict.get('freeze_first_only') or False,
                kernel=self.kernel_selector,
                uncertainty_type=self.model_config_selector_dict.get('uncertainty_type'),
                alpha=self.model_config_selector_dict.get('alpha'),
                n_jobs=self.n_jobs,
                seed=self.seed,
                logger=self.logger)
        return self._model_selector

    @property
    def model_evaluators(self):
        if not hasattr(self, '_model_evaluators'):
            self._model_evaluators = [get_model(
                data_format=model_config['data_format'],
                dataset_type=self.dataset_type,
                model=model_config.get('model'),
                save_dir='%s/evaluator_%d' % (self.save_dir, i),
                loss_function=model_config.get('loss_function'),
                num_tasks=len(self.target_columns),
                multiclass_num_classes=model_config.get('loss_function') or 3,
                features_generator=self.features_generator_evaluators[i],
                no_features_scaling=model_config.get('no_features_scaling') or False,
                features_only=model_config.get('features_only') or False,
                features_size=self.data_train_evaluators[i].features_size(),
                epochs=model_config.get('epochs') or 30,
                depth=model_config.get('depth') or 3,
                hidden_size=model_config.get('hidden_size') or 300,
                ffn_num_layers=model_config.get('ffn_num_layers') or 2,
                ffn_hidden_size=model_config.get('ffn_hidden_size'),
                dropout=model_config.get('dropout') or 0.0,
                batch_size=model_config.get('batch_size') or 50,
                ensemble_size=model_config.get('ensemble_size') or 1,
                number_of_molecules=model_config.get('number_of_molecules') or 1,
                mpn_shared=model_config.get('mpn_shared') or False,
                atom_messages=model_config.get('atom_messages') or False,
                undirected=model_config.get('undirected') or False,
                class_balance=model_config.get('class_balance') or False,
                checkpoint_dir=model_config.get('checkpoint_dir'),
                checkpoint_frzn=model_config.get('checkpoint_frzn'),
                frzn_ffn_layers=model_config.get('frzn_ffn_layers') or 0,
                freeze_first_only=model_config.get('freeze_first_only') or False,
                kernel=self.kernel_evaluators[i],
                uncertainty_type=model_config.get('uncertainty_type'),
                alpha=model_config.get('alpha'),
                n_jobs=self.n_jobs,
                seed=self.seed,
                logger=self.logger
            ) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._model_evaluators

    @property
    def data_train_selector(self):
        if not hasattr(self, '_data_train_selector'):
            self._data_train_selector = get_data(
                data_format=self.model_config_selector_dict['data_format'],
                path='%s/train_init.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_selector,
                features_combination=self.model_config_selector_dict.get('features_combination'),
                graph_kernel_type=self.model_config_selector_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_train_selector

    @property
    def data_pool_selector(self):
        if not hasattr(self, '_data_pool_selector'):
            self._data_pool_selector = get_data(
                data_format=self.model_config_selector_dict['data_format'],
                path='%s/pool_init.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_selector,
                features_combination=self.model_config_selector_dict.get('features_combination'),
                graph_kernel_type=self.model_config_selector_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_pool_selector

    @property
    def data_val_selector(self):
        if not hasattr(self, '_data_val_selector'):
            self._data_val_selector = get_data(
                data_format=self.model_config_selector_dict['data_format'],
                path='%s/val.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_selector,
                features_combination=self.model_config_selector_dict.get('features_combination'),
                graph_kernel_type=self.model_config_selector_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_val_selector

    @property
    def data_train_evaluators(self):
        if not hasattr(self, '_data_train_evaluators'):
            self._data_train_evaluators = [get_data(
                data_format=model_config['data_format'],
                path='%s/train_init.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluators[i],
                features_combination=model_config.get('features_combination'),
                graph_kernel_type=model_config.get('graph_kernel_type'),
                n_jobs=self.n_jobs) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._data_train_evaluators

    @property
    def data_pool_evaluators(self):
        if not hasattr(self, '_data_pool_evaluators'):
            self._data_pool_evaluators = [get_data(
                data_format=model_config['data_format'],
                path='%s/pool_init.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluators[i],
                features_combination=model_config.get('features_combination'),
                graph_kernel_type=model_config.get('graph_kernel_type'),
                n_jobs=self.n_jobs) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._data_pool_evaluators

    @property
    def data_val_evaluators(self):
        if not hasattr(self, '_data_val_evaluators'):
            self._data_val_evaluators = [get_data(
                data_format=model_config['data_format'],
                path='%s/val.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluators[i],
                features_combination=model_config.get('features_combination'),
                graph_kernel_type=model_config.get('graph_kernel_type'),
                n_jobs=self.n_jobs) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._data_val_evaluators

    @property
    def data_full_selector(self):
        if not hasattr(self, '_data_full_selector'):
            self._data_full_selector = get_data(
                data_format=self.model_config_selector_dict['data_format'],
                path=self.data_path,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_selector,
                features_combination=self.model_config_selector_dict.get('features_combination'),
                graph_kernel_type=self.model_config_selector_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_full_selector

    @property
    def data_full_evaluators(self) -> List:
        if not hasattr(self, '_data_full_evaluators'):
            self._data_full_evaluators = [get_data(
                data_format=model_config['data_format'],
                path=self.data_path,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluators[i],
                features_combination=model_config.get('features_combination'),
                graph_kernel_type=model_config.get('graph_kernel_type'),
                n_jobs=self.n_jobs) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._data_full_evaluators

    @property
    def features_generator_selector(self) -> Optional[List[FeaturesGenerator]]:
        fingerprints_class = self.model_config_selector_dict.get('fingerprints_class')
        radius = self.model_config_selector_dict.get('radius')
        num_bits = self.model_config_selector_dict.get('num_bits')
        if fingerprints_class is None:
            return None
        else:
            return [FeaturesGenerator(features_generator_name=fc,
                                      radius=radius,
                                      num_bits=num_bits) for fc in fingerprints_class]

    @property
    def features_generator_evaluators(self) -> Optional[List[List[FeaturesGenerator]]]:
        results = []
        for model_config in self.model_config_evaluators_dict:
            fingerprints_class = model_config.get('fingerprints_class')
            radius = model_config.get('radius')
            num_bits = model_config.get('num_bits')
            if fingerprints_class is None:
                results.append(None)
            else:
                results.append([FeaturesGenerator(features_generator_name=fc,
                                                  radius=radius,
                                                  num_bits=num_bits) for fc in fingerprints_class])
        return results

    @property
    def kernel_selector(self):
        return get_kernel(
            graph_kernel_type=self.model_config_selector_dict.get('graph_kernel_type'),
            mgk_files=self.model_config_selector_dict.get('mgk_files'),
            features_kernel_type=self.model_config_selector_dict.get('features_kernel_type'),
            features_hyperparameters=self.model_config_selector_dict.get('features_hyperparameters'),
            features_hyperparameters_file=self.model_config_selector_dict.get('features_hyperparameters_file'),
            dataset=self.data_full_selector,
            kernel_pkl_path='%s/kernel_selector.pkl' % self.save_dir,
        )

    @property
    def kernel_evaluators(self) -> List:
        if not hasattr(self, '_kernel_evaluators'):
            self._kernel_evaluators = [get_kernel(
                graph_kernel_type=model_config.get('graph_kernel_type'),
                mgk_files=model_config.get('mgk_files'),
                features_kernel_type=model_config.get('features_kernel_type'),
                features_hyperparameters=model_config.get('features_hyperparameters'),
                features_hyperparameters_file=model_config.get('features_hyperparameters_file'),
                dataset=self.data_full_evaluators[i],
                kernel_pkl_path='%s/kernel_evaluator_%d.pkl' % (self.save_dir, i),
            ) for i, model_config in enumerate(self.model_config_evaluators_dict)]
        return self._kernel_evaluators

    @property
    def selection_method(self):
        if not hasattr(self, '_selection_method'):
            if self.exploitive_target is not None and self.exploitive_target not in ['min', 'max']:
                self.exploitive_target = float(self.exploitive_target)
            if self.learning_type == 'passive':
                if self.batch_mode == 'cluster':
                    self._selection_method = ClusterRandomSelectionMethod(seed=self.seed)
                else:
                    self._selection_method = RandomSelectionMethod(seed=self.seed)
            elif self.learning_type == 'explorative':
                if self.batch_mode == 'cluster':
                    self._selection_method = ClusterExplorativeSelectionMethod(seed=self.seed)
                else:
                    self._selection_method = ExplorativeSelectionMethod(seed=self.seed)
            elif self.learning_type == 'exploitive':
                if self.batch_mode == 'cluster':
                    self._selection_method = ClusterExploitiveSelectionMethod(target=self.exploitive_target,
                                                                              seed=self.seed)
                else:
                    self._selection_method = ExploitiveSelectionMethod(target=self.exploitive_target, seed=self.seed)
            else:
                raise ValueError('Unknown learning type %s' % self.learning_type)
            self._selection_method.batch_size = self.batch_size
        return self._selection_method

    @property
    def forgetter(self):
        if not hasattr(self, '_forgetter'):
            if self.forget_protocol == 'forget_first':
                forgeter = FirstForgetter()
            elif self.forget_protocol == 'forget_random':
                forgeter = RandomForgetter(seed=0)
            elif self.forget_protocol == 'min_oob_uncertainty':
                forgeter = MinOOBUncertaintyForgetter(seed=0)
            elif self.forget_protocol == 'max_oob_uncertainty':
                forgeter = MaxOOBUncertaintyForgetter(seed=0)
            elif self.forget_protocol == 'min_oob_error':
                forgeter = MinOOBErrorForgetter(seed=0)
            elif self.forget_protocol == 'min_loo_error':
                forgeter = MinLOOErrorForgetter(seed=0)
            else:
                return None
            # set forget_size and forget_cutoff in forgetter.
            # get forget_size from forget_ratio
            if self.forget_ratio is not None:
                assert self.forget_size is None
                forgeter.forget_size = math.ceil(self.forget_ratio *
                                                 (len(self.data_train_selector) + len(self.data_pool_selector)))
            else:
                forgeter.forget_size = self.forget_size
            if self.forget_cutoff is not None:
                assert self.forget_size is None
                assert self.forget_protocol in ['max_oob_uncertainty', 'min_oob_uncertainty',
                                                'min_oob_error', 'min_loo_error']
            forgeter.forget_cutoff = self.forget_cutoff
            forgeter.batch_size = 1
            self._forgetter = forgeter
        return self._forgetter

    @property
    def top_k_id(self) -> Optional[List[int]]:
        if self.top_k is not None:
            assert 0. < self.top_k < 1., 'top_k must be in (0, 1).'
            assert len(self.target_columns) == 1
            n_top_k = math.ceil(self.top_k * (len(self.data_train_evaluators) + len(self.data_pool_selector)))
            y_AL = self.data_train_selector.y.ravel().tolist() + self.data_pool_selector.y.ravel().tolist()
            top_k_index = get_topn_idx(y_AL, n_top_k, target=self.exploitive_target)
            top_k_id = []
            for i, data in enumerate(self.data_train_selector.data + self.data_pool_selector.data):
                if i in top_k_index:
                    top_k_id.append(data.id)
            return top_k_id
        else:
            return None

    def process_args(self) -> None:
        super().process_args()
        # get stop_size from stop_ratio
        if self.stop_ratio is not None:
            if self.stop_size is None:
                self.stop_size = math.ceil(
                    self.stop_ratio * (len(self.data_train_selector) + len(self.data_pool_selector)))
            else:
                self.stop_size = min(
                    self.stop_size,
                    math.ceil(self.stop_ratio * (len(self.data_train_selector) + len(self.data_pool_selector))))
            assert self.stop_size >= 2
        # set the default maximum number of iterations of active learning
        if self.max_iter is None:
            self.max_iter = len(self.data_pool_selector)
        # check the input for exploitive active learning
        if self.learning_type == 'exploitive':
            assert self.dataset_type == 'regression', 'exploitive active learning only support regression task.'
            assert self.top_k is not None, 'top_k must be set for exploitive active learning.'
            assert self.exploitive_target is not None, 'exploitive_target must be set for exploitive active learning.'
        # check the input for forgetting active learning
        if self.forget_protocol is not None:
            assert (self.forgetter.forget_size is None) ^ (self.forgetter.forget_cutoff is None)
        # introduce error (Randomly flip the labels of a portion of the data) to training set
        if self.error_rate is not None:
            assert self.dataset_type == 'classification'
            assert 0. < self.error_rate <= 1.
            df = pd.read_csv('%s/train_init.csv' % self.save_dir)
            for i, data in enumerate(self.data_train_selector.data):
                if df.loc[i, 'flip_label'] is True:
                    data.flip_label = True
                else:
                    data.flip_label = False
            df = pd.read_csv('%s/pool_init.csv' % self.save_dir)
            for i, data in enumerate(self.data_pool_selector.data):
                if df.loc[i, 'flip_label'] is True:
                    data.flip_label = True
                else:
                    data.flip_label = False
        # check unique ID for the data sets.
        unique_id = []
        for data in self.data_train_selector.data + self.data_pool_selector.data:
            assert data.id not in unique_id
            unique_id.append(data.id)
        for data in self.data_val_selector:
            if self.full_val:
                assert data.id in unique_id
            else:
                assert data.id not in unique_id
                unique_id.append(data.id)

    def flip_labels(self, datasets, error_index):
        i = 0
        for dataset in datasets:
            for data in dataset:
                if i in error_index:
                    data.targets ^= 1
                    data.flip_label = True
                else:
                    data.flip_label = False
                i += 1


class ActiveLearningContinueArgs(CommonArgs):
    stop_ratio: float = None
    """the ratio of molecules to stop the active learning."""
    stop_size: int = None
    """the number of molecules to stop the active learning."""


class ReEvaluateArgs(CommonArgs):
    model_config_evaluator: str
    """config file contain all information of the machine learning model for performance evaluation."""
    evaluator_id: int
    """the output id of the evaluator"""
    evaluate_stride: int = 100
    """evaluate model performance on the validation set when the size of the training set is an integer multiple of the 
    evaluation stride."""
    metrics: List[Metric]
    """the metrics to evaluate model performance."""

    data_public = None
    """Use public data sets."""
    data_path: str = None
    """the Path of input data CSV file."""
    pure_columns: List[str] = None
    """
    For pure compounds.
    Name of the columns containing single SMILES or InChI string.
    """
    mixture_columns: List[str] = None
    """
    For mixtures.
    Name of the columns containing multiple SMILES or InChI string and 
    corresponding concentration.
    example: ['C', 0.5, 'CC', 0.3]
    """
    target_columns: List[str] = None
    """
    Name of the columns containing target values.
    """
    feature_columns: List[str] = None
    """
    Name of the columns containing additional features_mol such as temperature, 
    pressuer.
    """
    dataset_type: Literal['regression', 'classification', 'multiclass'] = None
    """
    Type of task.
    """
    full_val: bool = False
    """validate the performance of active learning on the full dataset."""

    @property
    def model_config_evaluator_dict(self) -> Dict:
        return json.loads(open(self.model_config_evaluator).read())

    @property
    def model_evaluator(self):
        if not hasattr(self, '_model_evaluator'):
            self._model_evaluator = get_model(
                data_format=self.model_config_evaluator_dict['data_format'],
                dataset_type=self.dataset_type,
                model=self.model_config_evaluator_dict.get('model'),
                save_dir='%s/evaluator_%d' % (self.save_dir, self.evaluator_id),
                loss_function=self.model_config_evaluator_dict.get('loss_function'),
                num_tasks=len(self.target_columns),
                multiclass_num_classes=self.model_config_evaluator_dict.get('loss_function') or 3,
                features_generator=self.features_generator_evaluator,
                no_features_scaling=self.model_config_evaluator_dict.get('no_features_scaling') or False,
                features_only=self.model_config_evaluator_dict.get('features_only') or False,
                features_size=self.data_al_evaluator.features_size(),
                epochs=self.model_config_evaluator_dict.get('epochs') or 30,
                depth=self.model_config_evaluator_dict.get('depth') or 3,
                hidden_size=self.model_config_evaluator_dict.get('hidden_size') or 300,
                ffn_num_layers=self.model_config_evaluator_dict.get('ffn_num_layers') or 2,
                ffn_hidden_size=self.model_config_evaluator_dict.get('ffn_hidden_size'),
                dropout=self.model_config_evaluator_dict.get('dropout') or 0.0,
                batch_size=self.model_config_evaluator_dict.get('batch_size') or 50,
                ensemble_size=self.model_config_evaluator_dict.get('ensemble_size') or 1,
                number_of_molecules=self.model_config_evaluator_dict.get('number_of_molecules') or 1,
                mpn_shared=self.model_config_evaluator_dict.get('mpn_shared') or False,
                atom_messages=self.model_config_evaluator_dict.get('atom_messages') or False,
                undirected=self.model_config_evaluator_dict.get('undirected') or False,
                class_balance=self.model_config_evaluator_dict.get('class_balance') or False,
                checkpoint_dir=self.model_config_evaluator_dict.get('checkpoint_dir'),
                checkpoint_frzn=self.model_config_evaluator_dict.get('checkpoint_frzn'),
                frzn_ffn_layers=self.model_config_evaluator_dict.get('frzn_ffn_layers') or 0,
                freeze_first_only=self.model_config_evaluator_dict.get('freeze_first_only') or False,
                kernel=self.kernel_evaluator,
                uncertainty_type=self.model_config_evaluator_dict.get('uncertainty_type'),
                alpha=self.model_config_evaluator_dict.get('alpha'),
                n_jobs=self.n_jobs,
                seed=self.seed,
                logger=self.logger
            )
        return self._model_evaluator

    @property
    def data_al_evaluator(self):
        if not hasattr(self, '_data_al_evaluator'):
            self._data_al_evaluator = get_data(
                data_format=self.model_config_evaluator_dict['data_format'],
                path='%s/train_al.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluator,
                features_combination=self.model_config_evaluator_dict.get('features_combination'),
                graph_kernel_type=self.model_config_evaluator_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_al_evaluator

    @property
    def data_val_evaluator(self):
        if not hasattr(self, '_data_val_evaluator'):
            self._data_val_evaluator = get_data(
                data_format=self.model_config_evaluator_dict['data_format'],
                path='%s/val.csv' % self.save_dir,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluator,
                features_combination=self.model_config_evaluator_dict.get('features_combination'),
                graph_kernel_type=self.model_config_evaluator_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_val_evaluator

    @property
    def data_full_evaluator(self):
        if not hasattr(self, '_data_full_evaluator'):
            self._data_full_evaluator = get_data(
                data_format=self.model_config_evaluator_dict['data_format'],
                path=self.data_path,
                pure_columns=self.pure_columns,
                mixture_columns=self.mixture_columns,
                target_columns=self.target_columns,
                feature_columns=self.feature_columns,
                features_generator=self.features_generator_evaluator,
                features_combination=self.model_config_evaluator_dict.get('features_combination'),
                graph_kernel_type=self.model_config_evaluator_dict.get('graph_kernel_type'),
                n_jobs=self.n_jobs)
        return self._data_full_evaluator

    @property
    def features_generator_evaluator(self) -> Optional[List[FeaturesGenerator]]:
        fingerprints_class = self.model_config_evaluator_dict.get('fingerprints_class')
        radius = self.model_config_evaluator_dict.get('radius')
        num_bits = self.model_config_evaluator_dict.get('num_bits')
        if fingerprints_class is None:
            return None
        else:
            return [FeaturesGenerator(features_generator_name=fc,
                                      radius=radius,
                                      num_bits=num_bits) for fc in fingerprints_class]

    @property
    def kernel_evaluator(self):
        return get_kernel(
            graph_kernel_type=self.model_config_evaluator_dict.get('graph_kernel_type'),
            mgk_files=self.model_config_evaluator_dict.get('mgk_files'),
            features_kernel_type=self.model_config_evaluator_dict.get('features_kernel_type'),
            features_hyperparameters=self.model_config_evaluator_dict.get('features_hyperparameters'),
            features_hyperparameters_file=self.model_config_evaluator_dict.get('features_hyperparameters_file'),
            dataset=self.data_full_evaluator,
            kernel_pkl_path='%s/kernel_evaluator_%d.pkl' % (self.save_dir, self.evaluator_id),
        )

    def process_args(self) -> None:
        super().process_args()
        if self.data_public == 'freesolv':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['freesolv']
            self.dataset_type = 'regression'
        elif self.data_public == 'delaney':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['logSolubility']
            self.dataset_type = 'regression'
        elif self.data_public == 'lipo':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['lipo']
            self.dataset_type = 'regression'
        elif self.data_public == 'pdbbind_refined':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['-logKd/Ki']
            self.dataset_type = 'regression'
        elif self.data_public == 'pdbbind_full':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['-logKd/Ki']
            self.dataset_type = 'regression'
        elif self.data_public in ['ld50_zhu', 'caco2_wang', 'solubility_aqsoldb', 'ppbr_az', 'vdss_lombardo',
                                  'Half_Life_Obach', 'Clearance_Hepatocyte_AZ']:
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['Drug']
            self.target_columns = ['Y']
            self.dataset_type = 'regression'
        elif self.data_public == 'bbbp':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['p_np']
            self.dataset_type = 'classification'
        elif self.data_public == 'bace':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['mol']
            self.target_columns = ['Class']
            self.dataset_type = 'classification'
        elif self.data_public == 'hiv':
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['smiles']
            self.target_columns = ['HIV_active']
            self.dataset_type = 'classification'
        elif self.data_public in ['ames', 'carcinogens_lagunin', 'dili', 'herg', 'skin', 'hia_hou', 'pgp_broccatelli',
                                  'bioavailability_ma', 'clintox', 'bbb_martins', 'CYP1A2_Veith',
                                  'CYP2C9_Substrate_CarbonMangels', 'CYP2C9_Veith', 'CYP2C19_Veith',
                                  'CYP2D6_Substrate_CarbonMangels', 'CYP2D6_Veith', 'CYP3A4_Veith',
                                  'CYP3A4_Substrate_CarbonMangels']:
            self.data_path = os.path.join(DATA_DIR, '%s.csv' % self.data_public)
            self.pure_columns = ['Drug']
            self.target_columns = ['Y']
            self.dataset_type = 'classification'
